!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  This file is AUTOMATICALLY GENERATED by:
!!  ../tools/fortran_wrappers.pl ../include/magmablas_zc.i
!!  Do not edit.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module magmablas_zcfortran

use magma_param

implicit none

!---- Fortran interfaces to MAGMA subroutines ----
interface

subroutine magmablasf_zcaxpycp( m, r, x, b, w )
    integer          :: m
    magma_devptr_t   :: r
    magma_devptr_t   :: x
    magma_devptr_t   :: b
    magma_devptr_t   :: w
end subroutine magmablasf_zcaxpycp

subroutine magmablasf_zclaswp( n, A, lda, SA, m, ipiv, incx )
    integer          :: n
    magma_devptr_t   :: A
    integer          :: lda
    magma_devptr_t   :: SA
    integer          :: m
    integer          :: ipiv(*)
    integer          :: incx
end subroutine magmablasf_zclaswp

subroutine magmablasf_zlag2c( m, n, A, lda, SA, ldsa, info )
    integer          :: m
    integer          :: n
    magma_devptr_t   :: A
    integer          :: lda
    magma_devptr_t   :: SA
    integer          :: ldsa
    integer          :: info
end subroutine magmablasf_zlag2c

subroutine magmablasf_clag2z( m, n, SA, ldsa, A, lda, info )
    integer          :: m
    integer          :: n
    magma_devptr_t   :: SA
    integer          :: ldsa
    magma_devptr_t   :: A
    integer          :: lda
    integer          :: info
end subroutine magmablasf_clag2z

subroutine magmablasf_zlat2c( uplo, n, A, lda, SA, ldsa, info )
    character        :: uplo
    integer          :: n
    magma_devptr_t   :: A
    integer          :: lda
    magma_devptr_t   :: SA
    integer          :: ldsa
    integer          :: info
end subroutine magmablasf_zlat2c

subroutine magmablasf_clat2z( uplo, n, SA, ldsa, A, lda, info )
    character        :: uplo
    integer          :: n
    magma_devptr_t   :: SA
    integer          :: ldsa
    magma_devptr_t   :: A
    integer          :: lda
    integer          :: info
end subroutine magmablasf_clat2z

end interface

end module magmablas_zcfortran
