#ifndef DTRSM_BATCHED_CONFIG_H
#define DTRSM_BATCHED_CONFIG_H

#define DTRSM_BATCHED_LEFT_NB2     2, 32
#define DTRSM_BATCHED_LEFT_NB4     4, 32
#define DTRSM_BATCHED_LEFT_NB8     8, 64
#define DTRSM_BATCHED_LEFT_NB16   16, 64
#define DTRSM_BATCHED_LEFT_NB32   32, 32

#define DTRSM_BATCHED_RIGHT_NB2    2, 64
#define DTRSM_BATCHED_RIGHT_NB4    4, 32
#define DTRSM_BATCHED_RIGHT_NB8    8, 16
#define DTRSM_BATCHED_RIGHT_NB16  16, 16
#define DTRSM_BATCHED_RIGHT_NB32  32, 64

#endif    // DTRSM_BATCHED_CONFIG_H