!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  This file is AUTOMATICALLY GENERATED by:
!!  ../tools/fortran_wrappers.pl ../include/magma_zc.i
!!  Do not edit.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module magma_zcfortran

use magma_param

implicit none

!---- Fortran interfaces to MAGMA subroutines ----
interface

subroutine magmaf_zcgesv_gpu( trans, n, nrhs, dA, ldda, ipiv, dipiv, dB, lddb, dX, lddx,  &
        dworkd, dworks, iter, info )
    character        :: trans
    integer          :: n
    integer          :: nrhs
    magma_devptr_t   :: dA
    integer          :: ldda
    integer          :: ipiv(*)
    magma_devptr_t   :: dipiv
    magma_devptr_t   :: dB
    integer          :: lddb
    magma_devptr_t   :: dX
    integer          :: lddx
    magma_devptr_t   :: dworkd
    magma_devptr_t   :: dworks
    integer          :: iter
    integer          :: info
end subroutine magmaf_zcgesv_gpu

subroutine magmaf_zcgetrs_gpu( trans, n, nrhs, dA, ldda, dipiv, dB, lddb, dX, lddx, dSX,  &
        info )
    character        :: trans
    integer          :: n
    integer          :: nrhs
    magma_devptr_t   :: dA
    integer          :: ldda
    magma_devptr_t   :: dipiv
    magma_devptr_t   :: dB
    integer          :: lddb
    magma_devptr_t   :: dX
    integer          :: lddx
    magma_devptr_t   :: dSX
    integer          :: info
end subroutine magmaf_zcgetrs_gpu

subroutine magmaf_zcposv_gpu( uplo, n, nrhs, dA, ldda, dB, lddb, dX, lddx, dworkd,  &
        dworks, iter, info )
    character        :: uplo
    integer          :: n
    integer          :: nrhs
    magma_devptr_t   :: dA
    integer          :: ldda
    magma_devptr_t   :: dB
    integer          :: lddb
    magma_devptr_t   :: dX
    integer          :: lddx
    magma_devptr_t   :: dworkd
    magma_devptr_t   :: dworks
    integer          :: iter
    integer          :: info
end subroutine magmaf_zcposv_gpu

subroutine magmaf_zcgeqrsv_gpu( m, n, nrhs, dA, ldda, dB, lddb, dX, lddx, iter, info )
    integer          :: m
    integer          :: n
    integer          :: nrhs
    magma_devptr_t   :: dA
    integer          :: ldda
    magma_devptr_t   :: dB
    integer          :: lddb
    magma_devptr_t   :: dX
    integer          :: lddx
    integer          :: iter
    integer          :: info
end subroutine magmaf_zcgeqrsv_gpu

end interface

end module magma_zcfortran
