#ifndef CTRSM_BATCHED_CONFIG_H
#define CTRSM_BATCHED_CONFIG_H

#define CTRSM_BATCHED_LEFT_NB2     2, 32
#define CTRSM_BATCHED_LEFT_NB4     4, 64
#define CTRSM_BATCHED_LEFT_NB8     8, 64
#define CTRSM_BATCHED_LEFT_NB16   16, 32
#define CTRSM_BATCHED_LEFT_NB32   32, 32

#define CTRSM_BATCHED_RIGHT_NB2    2, 32
#define CTRSM_BATCHED_RIGHT_NB4    4, 64
#define CTRSM_BATCHED_RIGHT_NB8    8, 64
#define CTRSM_BATCHED_RIGHT_NB16  16, 32
#define CTRSM_BATCHED_RIGHT_NB32  32, 32

#endif // CTRSM_BATCHED_CONFIG_H
