!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  This file is AUTOMATICALLY GENERATED by:
!!  ../tools/fortran_wrappers.pl ../include/magmablas_ds.i
!!  Do not edit.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module magmablas_dsfortran

use magma_param

implicit none

!---- Fortran interfaces to MAGMA subroutines ----
interface

subroutine magmablasf_dsaxpycp( m, r, x, b, w )
    integer          :: m
    magma_devptr_t   :: r
    magma_devptr_t   :: x
    magma_devptr_t   :: b
    magma_devptr_t   :: w
end subroutine magmablasf_dsaxpycp

subroutine magmablasf_dslaswp( n, A, lda, SA, m, ipiv, incx )
    integer          :: n
    magma_devptr_t   :: A
    integer          :: lda
    magma_devptr_t   :: SA
    integer          :: m
    integer          :: ipiv(*)
    integer          :: incx
end subroutine magmablasf_dslaswp

subroutine magmablasf_dlag2s( m, n, A, lda, SA, ldsa, info )
    integer          :: m
    integer          :: n
    magma_devptr_t   :: A
    integer          :: lda
    magma_devptr_t   :: SA
    integer          :: ldsa
    integer          :: info
end subroutine magmablasf_dlag2s

subroutine magmablasf_slag2d( m, n, SA, ldsa, A, lda, info )
    integer          :: m
    integer          :: n
    magma_devptr_t   :: SA
    integer          :: ldsa
    magma_devptr_t   :: A
    integer          :: lda
    integer          :: info
end subroutine magmablasf_slag2d

subroutine magmablasf_dlat2s( uplo, n, A, lda, SA, ldsa, info )
    character        :: uplo
    integer          :: n
    magma_devptr_t   :: A
    integer          :: lda
    magma_devptr_t   :: SA
    integer          :: ldsa
    integer          :: info
end subroutine magmablasf_dlat2s

subroutine magmablasf_slat2d( uplo, n, SA, ldsa, A, lda, info )
    character        :: uplo
    integer          :: n
    magma_devptr_t   :: SA
    integer          :: ldsa
    magma_devptr_t   :: A
    integer          :: lda
    integer          :: info
end subroutine magmablasf_slat2d

end interface

end module magmablas_dsfortran
